#include <Servo.h>
#define pin_base 3   // Servo Base
#define pin_L1 5     // Servo Hombro
#define pin_L2 6     // Servo Codo
#define pin_L3 10    // Servo Muñeca
#define pin_eff 9    // Servo Gripper
#define Trigger 13   // Trigger Sensor Ultrasonico
#define Echo 12      // Echo Sensor Ultrasonico

int motorhor = 7;    // HIGH para que la banda retroceda
int motorahor = 8;   // HIGH para que la banda avance
int motorpwm = 11;   // Regula la velocidad de la banda

Servo base;
Servo L1;
Servo L2;
Servo L3;
Servo eff;


float L = 1.5;          // Largo del objeto, Valor en centimetros 
float a = 30;           // Ancho del objeto; Valor en centimetros
float distancia = 15;   // Punto de recoleccion en relacion al sensor; Valor en centimetros
float velocidad = 150;  // Velocidad de la banda; Valor entre 150 - 255
int ser = 1;            // Print en el serial activado: 1; Desactivado: 0
float Soltado_X = 280;  // Posicion en la que soltara el objeto en X
float Soltado_Y = 0;    // Posicion en la que soltara el objeto en Y


int estado = 0;         // Detectando: 0; Recogiendo = 1; Soltando = 2
int count = 0;
double distance = 0;
float tfreno;
float dfreno;
float vbanda;

//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------- void setup -----------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void setup()
{
  // Servos
  base.attach(pin_base);
  L1.attach(pin_L1);
  L2.attach(pin_L2);
  L3.attach(pin_L3);
  eff.attach(pin_eff);
  
  // Sensor Ultrasonico
  pinMode (Trigger, OUTPUT);
  pinMode(Echo, INPUT);
  digitalWrite(Trigger, LOW);

  // Banda transportadora
  pinMode(motorhor, OUTPUT);
  pinMode(motorahor, OUTPUT);
  pinMode(motorpwm, OUTPUT);

  // Calculos
  vbanda = (0.0667 * velocidad - 5.50136); 
  tfreno = (distancia + L/2) / vbanda * 1000 - 900; // 900 como compensador del perfil de desaseleracion de la banda 
  dfreno = 430 - distancia*10;

  
  if (ser == 1) {
    Serial.begin(9600);
    
    Serial.print("||------ Robot MK2 Operativo -------||");
    Serial.println();
    Serial.print("Tiempo de freno [S]: ");
    Serial.println(tfreno);
    Serial.print("TVelocidad de la banda [cm/s]: ");
    Serial.println(vbanda);
  }
  axis_home();
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//--------------------- void loop ----------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void loop()
{
  switch (estado) {
    //--------------------------------------------------------
    //------------------ Iniciando la banda ------------------
    case 0:
      // Encender banda transportadora
      digitalWrite(motorhor, LOW);
      digitalWrite(motorahor, HIGH);
      analogWrite(motorpwm, velocidad);
      delay(2000);
      estado = 1;
      break;
      
    //--------------------------------------------------------
    //----------------- Detectando un objeto -----------------
    case 1:
      
      distance = Detectar();
      
      if (distance < 7){
        count ++;
        if (ser == 1) {
          Serial.print("count: ");
          Serial.println(count);
        }
        
      } else if (count > 0) {
        count --;
      }
      // Ignorando fantasmas con almenos 5 mediciones menores a la distancia escogido seguidas
      if (count == 5) {
        if (ser == 1){
          Serial.print("||-------- Objeto Detectado --------||");
          Serial.println();
          Serial.print("Distancia: ");
          Serial.print(distance);
          Serial.println();
        }
        count = 0;
        estado = 2;
      }
      break;
      
    //--------------------------------------------------------  
    //----------------- Recogiendo el objeto -----------------
    case 2: 
      delay(tfreno);
      
      if (ser == 1){
        Serial.print("||----- En Posicion de recoleccion -----||");
        Serial.println();
      }

      // Detener la banda
      digitalWrite(motorhor, LOW);
      digitalWrite(motorahor, LOW);
      delay(1000);

      // Posicion de recoleccion
      inversa(dfreno, 140, 95); 
      delay(1000);

      // Cerrando Gripper
      move_axis(5, 77);
      // Gripper(a); // Funcion no probada: poner el gripper hasta cierto ancho; Valor en milimetros 
      delay(1000);

      // Subiendo en la posicion de recoleccion
      inversa(dfreno, 140, 130);
      delay(1000);

      
      estado = 3;
      break;
      
    //--------------------------------------------------------  
    //------------------ Soltando el objeto ------------------
    case 3:
      // Posicion de soltado alta
      inversa(Soltado_X, Soltado_Y, 130);
      delay(1000);
      
      // Posicion de soltado baja
      inversa(Soltado_X, Soltado_Y, 100);
      delay(500);
      
      // Abriendo Gripper
      move_axis(5, 120);
      // Gripper(100); // Funcion no probada: poner el gripper hasta cierto ancho; Valor en milimetros 
      delay(500);

      // Volver a posicion de Home y reinicio del loop 
      axis_home();
      delay(1000);
      estado = 0;
      break;
  }
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//------ Deteccion de objetos con sensor ultrasonico ---------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
double Detectar() {
  long t;
  long d;

  digitalWrite(Trigger, HIGH);
  delayMicroseconds(10);
  digitalWrite(Trigger, LOW);

  t = pulseIn(Echo, HIGH);
  d = t / 59;
  if (ser == 1) {
      Serial.print("Distancia: ");
      Serial.print(d);
      Serial.println();
  }

  return d;
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------ Control Servos a Posicion Base ----------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void axis_home()
{
  inversa(241 , 0, 170); // Posicion Home
  move_axis(5, 120);
  if (ser ==1) {
    Serial.print("||--- Servos en posicion inicial ---||");
    Serial.println();
    }
  delay(100);
}

//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//-------- Mover el gripper segun el ancho deseado -----------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void Gripper(float ancho)
{
  float Lg;
  float q;
  float a;
  
  Lg = 59.43; // Largo del gripper 
  a = 7.144; // Ancho del agarre
  q = asin((a + ancho/2 - 10)/Lg)* 4068/71 + 67.28;
  move_axis( 5, q);
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//-------------- Control Servos ------------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void move_axis(int servoId, float position)
{
  if (servoId == 1) // Servo Base
  { base.write(position);
    delay(100);
  }
  //--------------------------------------------------------
  if (servoId == 2) // Servo Hombro
  { L1.write(position);
    delay(100);
  }
  //--------------------------------------------------------
  if (servoId == 3) // Servo Codo
  { L2.write(position);
    delay(100);
  }
  //--------------------------------------------------------
  if (servoId == 4) // Servo Muñeca
  { L3.write(position);
    delay(100);
  }
  //--------------------------------------------------------
  if (servoId == 5) // Servo Gripper
  { eff.write(position);
    delay(100);
  }
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//-------------- Cinematica Inversa --------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
void inversa(float x1, float y1, float z1)
{
  // Largo de los links del brazo
  float l1 = 135; // Largo entre el eje del codo y del hombro
  float l2 = 147; // Largo del antebrazo
  float l3 = 70;  // Largo entre el fin del antebrazo y el eje de la muñeca
  float l4 = 24;  // Largo desde el eje de la muñeca y la posicion del gripper
  float l5 = 59;  // Altura del gripper segun donde se quiera agarrar el objeto (Hacerlo variar en caso de que se quiera variar mucho el ancho del objeto agarrado) 
  float z0 = 94;  // Altura desde el piso hasta el servo del Hombro

  // Desfase de X y Z dadas las dimensiones del gripper en caso de que se quiera agarrar siempre perpendicular a la banda 
  float x = x1 - l4;
  float y = y1;
  float z = z1 + l5;

  // Calculo de la inversa del brazo
  float q0_rad = atan(y / x);
  float q2_rad = acos((pow((l1 * cos(q0_rad)), 2) + pow((l2 * cos(q0_rad)), 2) - pow((x - l3 * cos(q0_rad)), 2) - pow((z - z0), 2)) / (2 * l1 * l2 * pow((cos(q0_rad)), 2)));
  float q1_rad = atan((x - l3 * cos(q0_rad)) / (z - z0)) - acos((pow((l1 * cos(q0_rad)), 2) - pow((l2 * cos(q0_rad)), 2) + pow((x - l3 * cos(q0_rad)), 2) + pow((z - z0), 2)) / (2 * l1 * cos(q0_rad) * sqrt(pow((x - l3 * cos(q0_rad)), 2) + pow((z - z0), 2))));
  float q3_rad = 3.14 / 2 + q0_rad;

  // Paso desde radianes a grados de los angulos 
  float q0 = q0_rad * 4068 / 71;
  float q1 = q1_rad * 4068 / 71;
  float q2 = q2_rad * 4068 / 71;
  float q3 = q3_rad * 4068 / 71 - 10;

  if (ser ==1) {
    Serial.print("Servo Base:   ");
    Serial.print(90 - 2 * q0);
    Serial.println();
    Serial.print("Servo Hombro: ");
    Serial.print(90 + q1);
    Serial.println();
    Serial.print("Servo Codo:   ");
    Serial.print(q2 - q1);
    Serial.println();
    Serial.print("Servo Muñeca: ");
    Serial.print(q3);
    Serial.println();
    Serial.println();
  }
  // Mover los servos segun los qi calculados para la configuracion de los servos
  move_axis( 1, 90 - 2 * q0 );
  move_axis( 2, 90 + q1);
  move_axis( 3, q2 - q1);
  move_axis( 4, q3);
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
//-------------- Cinematica Directa --------------------------------------------------//
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
// Funcion si es que desease monitorear las posiciones a las que estaria llegando el robot 
void directa(int q1, int q2, int q3, int q4, int q5)
{
  
}
//------------------------------------------------------------------------------------//
//------------------------------------------------------------------------------------//
